/* mp3dup.c
 * - mp3dup program
 * Copyright (c) 2002 Alexander Havng
 *
 * This program is free software; you can do whatever you want with
 * it except if your name is Ulbrecht.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <openssl/evp.h>
#include <glib.h>
#include <errno.h>

/* Note on comparisons
   Given file X and Y, the following cases exist
A)   X is not the same song as Y
B)   X is the same song as Y
C)   X is cut off, and thus the first N bytes of X are the same as the first N bytes of Y
D)   X is a portion of Y, where Y + offset = X
E)   X is a portion of, and cut off version off Y, where Y + offset = X for the first N bytes.

A and B are trivial.

C is solved by hashing on the md5sum of the first N bytes of X and Y and then comparing
D is solved by hashing on the md5sum of the last N bytes of X and Y

E is a pain in the ass
*/

typedef struct mp3dup_size_lookup_St {
	 gchar *filename;
	 GHashTable *table;
} mp3dup_size_lookup_t;

typedef struct mp3dup_md5_lookup_St {
	 GList *list;
} mp3dup_md5_lookup_t;

typedef struct mp3dup_lookup_St {
	 gchar *filename;
	 gint64 filesize;
} mp3dup_lookup_t;

static gint64 mp3dup_md5sums = 0, mp3dup_dirs = 0, mp3dup_irregulars = 0,
	 mp3dup_totalsize = 0, mp3dup_size_hits = 0, mp3dup_md5_hits = 0,
	 mp3dup_stat_hits = 0, mp3dup_io_read = 0, mp3dup_files = 0, mp3dup_dups = 0;

static gint mp3dup_minimum_portion_size = 4096;
static gboolean mp3dup_check_portions = FALSE;
static gboolean mp3dup_check_id3 = FALSE;
static gboolean mp3dup_silent = FALSE;

static void mp3dup_check_dir (const gchar *dir, GHashTable *size_table);
static gint mp3dup_md5sum (const gchar *filename, guint8 *md5sum);
static void mp3dup_check_file (const gchar *filename, GHashTable *size_table);
static gint64 mp3dup_check_duplicate (const gchar *file, const gchar *file2);
static void mp3dup_stats (GHashTable *table, GTimer *timer);
static void mp3dup_head_and_tail_compare (const gchar *filename, gint64 filesize,
										  gint minimum_portion_size);
static gint64 mp3dup_check_bytechunk (const guint8 *buf, const guint8 *buf2, guint size, guint size2, gint64 equals);
static void mp3dup_id3_check (const gchar *filename, gint64 filesize);

void
mp3dup_usage (void)
{
	 fprintf (stderr, "Usage: mp3dup [-p portionbytes] [-i] [-s] <path> <path> <path>\n");
}

gint
main (gint argc, gchar **argv)
{
	 GHashTable *size_table;
	 gint i = 1, k;
	 GTimer *timer = g_timer_new ();
	 
	 if (argc < 2) {
		  mp3dup_usage ();
		  exit (1);
	 }

	 size_table = g_hash_table_new (NULL, NULL);
	 if (!size_table) {
		  fprintf (stderr, "Could not create hash table, exiting\n");
		  return 1;
	 }

	 g_timer_start (timer);

	 do {
		  if (argv[i][0] == '-') {
			   for (k = 1; (k > 0) && argv[i][k]; k++) {
					switch (argv[i][k]) {
						 case 'p':
							  mp3dup_minimum_portion_size = atoi (argv[i + 1]);
							  mp3dup_check_portions = TRUE;
							  k = -1;
							  i++;
							  break;
						 case 'i':
							  mp3dup_check_id3 = TRUE;
							  break;
						 case 's':
							  mp3dup_silent = TRUE;
							  break;
						 default:
							  mp3dup_usage ();
							  exit (0);
							  break;
					}
			   }
		  } else {
			   mp3dup_check_dir (argv[i], size_table);
		  }

		  i++;
		  
	 } while (i < argc);
	 
	 g_timer_stop (timer);

	 mp3dup_stats (size_table, timer);
	 
	 return 0;
}

static void
mp3dup_stats (GHashTable *table, GTimer *timer)
{
	 if (mp3dup_silent)
		  return;
	 
	 fprintf (stderr, "Duplicate check completed, %u different file sizes encountered\n", g_hash_table_size (table));
	 fprintf (stderr, "Checked %lld directories with %lld files, total %lld bytes of data\n", mp3dup_dirs, mp3dup_files, mp3dup_totalsize);
	 fprintf (stderr, "Found %lld irregular or unreadable files\n", mp3dup_irregulars);
	 fprintf (stderr, "%lld files had the same size, %lld files had the same md5sum, and %lld files were duplicates\n", mp3dup_size_hits, mp3dup_md5_hits, mp3dup_dups);
	 fprintf (stderr, "Complete i/o reads was %lld bytes\n", mp3dup_io_read);
	 fprintf (stderr, "Time elapsed was %2.2f seconds, %2.4f seconds per file\n", g_timer_elapsed (timer, NULL), g_timer_elapsed (timer, NULL) / (double)mp3dup_files);
}

static void
mp3dup_check_dir (const gchar *dir, GHashTable *size_table)
{
	DIR *d = opendir (dir);
	struct dirent dnt_e, *dnt = NULL;
	gchar *name;
	struct stat st;
	
	if (!d) {
		 fprintf (stderr, "ERROR: Could not open directory %s [%s], skipping\n", dir, strerror (errno));
		 return;
	}

	mp3dup_dirs++;
	
	while (readdir_r (d, &dnt_e, &dnt) == 0) {
		if (!dnt)
			break;
		if ((strcmp (dnt->d_name, ".") == 0)
			|| (strcmp (dnt->d_name, "..") == 0))
			 continue;
		name = g_strdup_printf ("%s%c%s", dir, G_DIR_SEPARATOR, dnt->d_name);

		if (stat (name, &st) != 0) {
			 fprintf (stderr, "ERROR: Could not stat %s [%s], skipping\n", name, strerror (errno));
			 g_free (name);
			 continue;
		}

		if (S_ISREG (st.st_mode)) {
			 mp3dup_check_file (name, size_table);
		} else if (S_ISDIR (st.st_mode)) {
			 mp3dup_check_dir (name, size_table);
		} else {
			 mp3dup_irregulars++;
		}
		
		g_free (name);
	}
	
	closedir (d);
}

static int
mp3dup_md5sum (const gchar *filename, guint8 *md5sum)
{
	MD5_CTX c;
	guint8 buf[4096];
	gint fd;
	gint64 size;

	mp3dup_md5sums++;
	
	memset (md5sum, 0, 34);
	
	if ((fd = open (filename, O_RDONLY)) == -1) {
		fprintf (stderr, "ERROR: Could not get md5sum for file %s [%s]\n", filename, strerror (errno));
		return 0;
	}

	MD5_Init (&c);
	
	while ((size = read (fd, buf, 4096)) > 0) {
		 mp3dup_io_read += size;
		 MD5_Update (&c, buf, size);
	}

	MD5_Final (md5sum, &c);
	close (fd);

	return 1;
}

static void
mp3dup_check_file (const gchar *filename, GHashTable *size_table)
{
	 guint8 md5sum[40];
	 struct stat st;
	 GList *node;
	 mp3dup_md5_lookup_t *lookup;
	 mp3dup_size_lookup_t *size_lookup;
	 gint64 filesize;

	 if (stat (filename, &st) != 0) {
		  fprintf (stderr, "ERROR: Could not stat %s [%s], skipping\n", filename, strerror (errno));
		  return;
	 }

	 mp3dup_files++;
	 filesize = st.st_size;

	 if (filesize == 0) {
		fprintf (stderr, "EMPTY: Skipping empty file %s\n", filename);
		return;
	 }

	 mp3dup_totalsize += filesize;

	 if (mp3dup_check_portions)
		  mp3dup_head_and_tail_compare (filename, filesize, mp3dup_minimum_portion_size);

	 size_lookup = g_hash_table_lookup (size_table, GUINT_TO_POINTER((guint)filesize));

	 if (size_lookup) {
		  mp3dup_size_hits++;

		  /* Filename means that no md5sum has been run */
		  if (size_lookup->filename) {
			   lookup = g_new0 (mp3dup_md5_lookup_t, 1);
			   lookup->list = g_list_append (lookup->list, size_lookup->filename);
			   if (mp3dup_md5sum (size_lookup->filename, md5sum) != 1) {
					fprintf (stderr, "ERROR: Could not md5sum %s, skipping\n", filename);
					return;
			   }
			   size_lookup->filename = NULL;
			   size_lookup->table = g_hash_table_new (g_str_hash, g_str_equal);
			   g_hash_table_insert (size_lookup->table, g_strdup (md5sum), lookup);
		  }
		  
		  if (mp3dup_md5sum (filename, md5sum) != 1) {
			   fprintf (stderr, "ERROR: Could not md5sum %s, skipping\n", filename);
			   return;
		  }
		  
		  lookup = g_hash_table_lookup (size_lookup->table, md5sum);
		  
		  if (lookup) {
			   mp3dup_md5_hits++;
			   
			   /* Same md5sum */
			   for (node = lookup->list; node; node = g_list_next (node)) {
					if (stat (node->data, &st) == 0) {
						 if (st.st_size == filesize) {
							  mp3dup_stat_hits++;
							  if (mp3dup_check_duplicate (node->data, filename) == 0) {
								   	fprintf (stdout, "Match:Identical:%s:%s\n", filename,
											 (gchar *)node->data);
									mp3dup_dups++;
									return;
							  }
						 } 
					} else {
						 fprintf (stderr, "ERROR: Could not stat %s [%s], skipping\n", (gchar *)node->data,
								  strerror (errno));
						 continue;
					}
			   }
			   
			   /* Same md5sum, but no match */
			   lookup->list = g_list_prepend (lookup->list, g_strdup (filename));
		  } else {
			   /* Different md5sum */
			   lookup = g_new0 (mp3dup_md5_lookup_t, 1);
			   lookup->list = g_list_append (lookup->list, g_strdup (filename));
			   g_hash_table_insert (size_lookup->table, g_strdup (md5sum), lookup);

		  }
	 } else {
		  size_lookup = g_new0 (mp3dup_size_lookup_t, 1);
		  size_lookup->filename = g_strdup (filename);
		  g_hash_table_insert (size_table, GUINT_TO_POINTER ((guint)filesize), size_lookup);
	 }

	 if (mp3dup_check_id3)
		  mp3dup_id3_check (filename, filesize);
}

static gint64
mp3dup_check_bytechunk (const guint8 *buf, const guint8 *buf2, guint size, guint size2, gint64 equals)
{
	 register gint i = 0;

	 for (i = 0; (i < size) && (i < size2); i++) {
		  if (buf[i] != buf2[i])
			   return (gint64)i + equals;
	 }
	 
	 return (gint64)i + equals;
}

static gint64
mp3dup_check_duplicate (const gchar *file, const gchar *file2)
{
	gint fd, fd2;
	gint size, size2;
	gint64 equals = 0;
	guint8 buf[4096], buf2[4096];
	
	if ((fd = open (file, O_RDONLY)) == -1) {
		 fprintf (stderr, "ERROR: Could not open %s [%s] for comparison\n", file, strerror (errno));
		 return -1;
	} else if ((fd2 = open (file2, O_RDONLY)) == -1) {
		 fprintf (stderr, "ERROR: Could not open %s [%s] for comparison\n", file2, strerror (errno));
		 return -1;
	}

	do {
		 size = read (fd, buf, 4096);
		 size2 = read (fd2, buf2, 4096);

		 mp3dup_io_read += size + size2;
		 
		 if ((size != size2) || ((size > 0) && (memcmp (buf, buf2, size) != 0))) {
			  /* Need to check these two bytechunks too */
			  equals = mp3dup_check_bytechunk (buf, buf2, size, size2, equals);
			  close (fd);
			  close (fd2);
			  if (equals > 0) {
				   return equals;
			  }
			  return -1;
		 }
		 
		 equals += size;
				 
	} while ((size > 0) && (size2 > 0));
	
	close (fd);
	close (fd2);

	return 0;	 
}

static gboolean
mp3dup_calculate_head_and_tailsums (const gchar *filename, gint64 filesize,
									guint8 *headsum, guint8 *tailsum)
{
	 gint fd;
	 gint64 size;
	 MD5_CTX c1, c2;
	 guint8 buf[4096];
	 
	 if ((fd = open (filename, O_RDONLY)) == -1) {
		  fprintf (stderr, "ERROR: Could not open %s [%s] for head and tail comparison\n", filename, strerror (errno));
		  return FALSE;
	 }

	 MD5_Init (&c1);
	 MD5_Init (&c2);

	 memset (headsum, 0, 34);
	 memset (tailsum, 0, 34);
	 
	 size = read (fd, buf, 4096);
	 
	 if (size > 0) {
		  mp3dup_io_read += size;
		  MD5_Update (&c1, buf, size);
		  MD5_Final (headsum, &c1);

		  if (filesize > 4096) {
			   lseek (fd, filesize - 4096, SEEK_SET);
			   size = read (fd, buf, 4096);
			   if (size > 0) {
					MD5_Update (&c2, buf, size);
					MD5_Final (tailsum, &c2);
			   }
		  } else {
			   memcpy (tailsum, headsum, 34);
		  }
	 } else {
		  fprintf (stderr, "ERROR: Could not read from file %s [%s]\n", filename, strerror (errno));
		  close (fd);
		  return FALSE;
	 }

	 close (fd);
	 
	 return TRUE;
}

static void
mp3dup_compare_heads (const gchar *filename, mp3dup_md5_lookup_t *lookup, gint minimum_portion_size)
{
	 GList *node;
	 gchar *comparefile;
	 gint64 equals;

	 for (node = lookup->list; node; node = g_list_next (node)) {
		  comparefile = node->data;
		  
		  equals = mp3dup_check_duplicate (filename, comparefile);
		  if (equals > minimum_portion_size) { /* Full match is 0 */
			   fprintf (stdout, "Match:Head-Partial %lld:%s:%s\n", equals, filename, comparefile);
		  }
	 }
}

static void
mp3dup_compare_tails (const gchar *filename, gint64 filesize, mp3dup_md5_lookup_t *lookup, gint minimum_portion_size)
{
	 GList *node;
	 gchar *comparefile;
	 gint fd, fd2, size, size2;
	 guint8 buf[4096], buf2[4096];
	 struct stat st;
	 gboolean match;
	 
	 if ((fd = open (filename, O_RDONLY)) == -1) {
		  fprintf (stderr, "ERROR: Could not open %s [%s] for tail comparison\n", filename, strerror (errno));
		  return;
	 }

	 for (node = lookup->list; node; node = g_list_next (node)) {
		  comparefile = node->data;
		  lseek (fd, 0, SEEK_SET);
		  
		  if ((fd2 = open (comparefile, O_RDONLY)) == -1) {
			   fprintf (stderr, "ERROR: Could not open %s [%s] for tail comparison\n", comparefile, strerror (errno));
			   continue;
		  }

		  if (fstat (fd2, &st) != 0) {
			   fprintf (stderr, "ERROR: Could not stat %s [%s] for tail comparison\n", comparefile, strerror (errno));
			   close (fd2);
			   continue;
		  }

		  if (st.st_size == filesize) { /* Will be identical match later */
			   close (fd2);
			   continue;
		  }

		  if (st.st_size > filesize) {
			   if (st.st_size - (st.st_size - filesize) < minimum_portion_size) {
					close (fd2);
					continue;
			   }
			   lseek (fd2, st.st_size - filesize, SEEK_SET);
		  } else {
			   if (filesize - (filesize - st.st_size) < minimum_portion_size) {
					close (fd2);
					continue;
			   }
			   lseek (fd, filesize - st.st_size, SEEK_SET);
		  }

		  match = TRUE;

		  do {
			   size = read (fd, buf, 4096); 
			   size2 = read (fd2, buf2, 4096);

			   mp3dup_io_read += size + size2;
			   
			   if ((size != size2) || (memcmp (buf, buf2, size) != 0)) {
					match = FALSE;
					break;
			   }
		  } while ((size > 0));

		  close (fd2);
		  if (match) {
			   fprintf (stdout, "Match:Tail-Partial %lld:%s:%s\n", filesize, filename, comparefile);
		  }
	 }
	 
	 close (fd);
}

static void
mp3dup_head_and_tail_compare (const gchar *filename, gint64 filesize,
							  gint minimum_portion_size)
{
	 static GHashTable *head_table = NULL, *tail_table = NULL;
	 guint8 headsum[40], tailsum[40];
	 mp3dup_md5_lookup_t *lookup;

	 if (!head_table) {
		  head_table = g_hash_table_new (g_str_hash, g_str_equal);
		  tail_table = g_hash_table_new (g_str_hash, g_str_equal);
	 }

	 if (!mp3dup_calculate_head_and_tailsums (filename, filesize, headsum, tailsum))
		  return;

	 if (filesize < minimum_portion_size)
		  return;
	 
	 /* Look for C */
	 lookup = g_hash_table_lookup (head_table, headsum);

	 if (lookup) {
		  mp3dup_compare_heads (filename, lookup, minimum_portion_size);
		  lookup->list = g_list_prepend (lookup->list, g_strdup (filename));
	 } else {
		  lookup = g_new0 (mp3dup_md5_lookup_t, 1);
		  lookup->list = g_list_append (lookup->list, g_strdup (filename));
		  g_hash_table_insert (head_table, g_strdup (headsum), lookup);
	 }
	 
	 /* Look for D */
	 lookup = g_hash_table_lookup (tail_table, tailsum);

	 if (lookup) {
		  mp3dup_compare_tails (filename, filesize, lookup, minimum_portion_size);
		  lookup->list = g_list_prepend (lookup->list, g_strdup (filename));
	 } else {
		  lookup = g_new0 (mp3dup_md5_lookup_t, 1);
		  lookup->list = g_list_append (lookup->list, g_strdup (filename));
		  g_hash_table_insert (tail_table, g_strdup (tailsum), lookup);
	 }
}

static void
mp3dup_id3_compare (const gchar *filename, const gchar *id3buf, const gchar *comparefile)
{
	 gint fd;
	 gchar buf[128];
	 
	 if ((fd = open (comparefile, O_RDONLY)) == -1) {
		  fprintf (stderr, "ERROR: Could not open %s [%s] for id3 compare\n", comparefile, strerror (errno));
		  return;
	 }

	 if (lseek (fd, -128, SEEK_END) == -1) { /* Weird, cause we read it before */
		  close (fd);
		  return;
	 }

	 if (read (fd, buf, 128) == 128) { /* Dito weird */
		  if (strncmp (buf, id3buf, 128) == 0) {
			   fprintf (stdout, "Match:ID3:%s:%s\n", filename, comparefile);
		  }
	 }

	 close (fd);
}

static void
mp3dup_id3_check (const gchar *filename, gint64 filesize)
{
	 gint fd;
	 static GHashTable *tag_table = NULL;
	 gchar buf[128];
	 guint8 md5sum[40];
	 MD5_CTX c;
	 mp3dup_md5_lookup_t *lookup;
	 
	 if (!tag_table) {
		  tag_table = g_hash_table_new (g_str_hash, g_str_equal);
	 }

	 if (filesize <= 128)
		  return;

	 if ((fd = open (filename, O_RDONLY)) == -1) {
		  fprintf (stderr, "ERROR: Could not open %s [%s] for id3 comparison\n", filename, strerror (errno));
		  return;
	 }

	 if (lseek (fd, -128, SEEK_END) == -1) {
		  fprintf (stderr, "ERROR: Could not position file %s [%s] for id3 tag\n", filename, strerror (errno));
		  return;
	 }

	 if (read (fd, buf, 128) != 128) {
		  close (fd);
		  return;
	 }

	 mp3dup_io_read += 128;
		  
	 if (strncmp (buf, "TAG", 3) == 0) {
		  mp3dup_md5sums++;
		  memset (md5sum, 0, 34);
		  MD5_Init (&c);
		  MD5_Update (&c, buf, 128);
		  MD5_Final (md5sum, &c);

		  lookup = g_hash_table_lookup (tag_table, md5sum);
		  if (lookup) {
			   GList *node;
			   for (node = lookup->list; node; node = g_list_next (node)) {
					mp3dup_id3_compare (filename, buf, node->data);
			   }
			   lookup->list = g_list_prepend (lookup->list, g_strdup (filename));
		  } else {
			   lookup = g_new0 (mp3dup_md5_lookup_t, 1);
			   lookup->list = g_list_append (lookup->list, g_strdup (filename));
			   g_hash_table_insert (tag_table, g_strdup (md5sum), lookup);
		  }
	 }

	 close (fd);
}




